/*
 * $RCSfile: process.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993,1991,1990,1996  Arizona Board of Regents
 *
 * $Log: process.c,v $
 * Revision 1.8  1997/07/01 18:57:53  mjk
 * Make all threads PTHREAD_DETACHED
 *
 * Revision 1.7  1997/03/10 21:37:34  mjk
 * *** empty log message ***
 *
 * Revision 1.6  1996/11/22 18:42:23  mjk
 * *** empty log message ***
 *
 * Revision 1.4  1996/05/31 21:45:39  mjk
 * Set the stack size for events.
 *
 * Revision 1.3  1996/03/15  22:39:24  mjk
 * timespec is now POSIX
 *
 * Revision 1.2  1996/03/11  18:02:23  mjk
 * Udptest passes w/ Linux 1.2.13 and Pthreads.  All signals are caught by
 * specific thread(s).
 *
 * Revision 1.1  1996/02/19  20:21:02  mjk
 * Initial revision
 *
 */

#include <stdio.h>
#include <stdarg.h>
#include <errno.h>
#include <limits.h>
#include <sys/time.h>
#include "config.h"
#include "thread.h"
#include "semaphore.h"
#include "xtype.h"
#include "xkernel.h"	
#include "event_i.h"

/* externs */
void bufferPoolInit(void);
void xkBufferPoolInit(void);
void CreateKernelProcess(CreateProcFunc, int, int, int, int);
void Yield(void);
void inputThreadInit( void );

static void *enter_master_sledgehammer_monitor(void *);
static void *signal_handler_thread(void *);

#define MAXARGS 6
#define MAX_KERNEL_THREADS_FOR_THREADS 5
#define XK_MAX_RUNNING_THREADS 24	/* try to keep running threads less than this */

static int current_xk_threads = 0;
int max_xk_threads_inuse = 0;
int xk_thread_limit = MAX_KERNEL_THREADS_FOR_THREADS;


pthread_mutex_t sledgehammer_concurrency_control;
#define XTHREAD_SELF()	pthread_self()
#define XTHREAD_TYPE	pthread_t

#ifdef XKLOCKDEBUG
int xklockdepthreq = 0;
int xklockdepth = 0;
int tracexklock = 0;
#endif XKLOCKDEBUG


/*
 * xk_master_lock, xk_master_unlock
 *
 * user callable master concurrency control
 *
 */
void
xk_master_lock(void)
{
    MASTER_LOCK;
}

void
xk_master_unlock(void)
{
    MASTER_UNLOCK;
}

void semWait(Semaphore *s)
{
  xTrace1(processswitch, TR_GROSS_EVENTS, "semWait: %x", s);
  xk_master_unlock();
  sema_wait(s);
  xk_master_lock();
}

void semSignal(Semaphore *s)
{
  xTrace1(processswitch, TR_GROSS_EVENTS, "semSignal: %x", s);
  sema_post(s);
}

void semInit(Semaphore *s,unsigned n) 
{
  sema_init(s,0,n);
}

/*
 *
 * threadInit
 *
 */

pthread_attr_t	thread_attributes;

void threadInit(void)
{
  pthread_t	tid;
  
  xTrace0(processcreation, TR_FULL_TRACE, "Enter event function threadInit");

				/* Start the signal handler thread.
                                   Before this is created relevant
                                   signals should be blocked by the
                                   process. */
  pthread_attr_init(&thread_attributes);
  pthread_attr_setdetachstate(&thread_attributes, 1);
  pthread_create(&tid, &thread_attributes, signal_handler_thread, NULL);
  
  pthread_mutex_init(&sledgehammer_concurrency_control, 0);
}

void inputThreadInit(void)
{
  /* defined in input_proc.c */
  /*  xkFillThreadPool(); */
  xkBufferPoolInit();
}

static void xkernel_activate_thread(int *, int);

/*
 *  xkernel_activate_thread
 *
 *     startup a thread, given the arguments
 *
 */
static void xkernel_activate_thread(int *args, int priority)
{
  XTHREAD_TYPE		child;
  struct sched_param	param;
  int 			thr_status = 0;
    
  xTrace0(processcreation, TR_FULL_TRACE, "xkernel_activate_thread enter");

  pthread_attr_getschedparam(&thread_attributes, &param);
  param.sched_priority = priority;	/* NOT POSIX 				*/
  pthread_attr_setschedparam(&thread_attributes, &param);
  pthread_attr_setstacksize(&thread_attributes, EVENT_STACK_SIZE);  
  thr_status = pthread_create(&child, &thread_attributes,
			      enter_master_sledgehammer_monitor, args);

  if (!thr_status) {
    xTrace1(processcreation, TR_MAJOR_EVENTS,
	    "Created thread_id: %d", child);
    xTrace1(processswitch, TR_GROSS_EVENTS,
	    "Created thread_id: %d", child);
  }
  else {
    xTrace1(processswitch, TR_ERRORS,
	    "xkernel_activate_thread: pthread_create failed (%d)",
	    thr_status);
  }
}

/*
 *
 * CreateProcess(function_ptr, priority, argcount, arg1, arg2, ...)
 *
 */

bool
CreateProcess(void *r, int pri, int nargs, ... )
{
  va_list ap;
  short prio;
  int i;
  int *arg_vec;
  int *argp;

  xTrace0(processcreation, TR_EVENTS, "Enter CreateProcess");

  va_start(ap, nargs);
  prio = pri;
 
  arg_vec = (int *)xMalloc((MAXARGS+2)*sizeof(int));
  argp = arg_vec;
  *argp++ = (int)r;
  *argp++ = nargs;
  for (i=0; i<nargs; i++)
    *argp++ = va_arg(ap, int);

  xTrace3(processcreation, TR_MAJOR_EVENTS,
	  "pthreads CreateProcess: func_addr %x nargs %d  pri %d",
	  r, nargs, pri);
  xkernel_activate_thread(arg_vec, prio);
  return(TRUE);
}

/*
 *  CreateKernelProcess
 *
 *
 */
void
CreateKernelProcess(CreateProcFunc r, int prio, int arg1, int arg2, int arg3)
{
  CreateProcess((void *)r, prio, 3, arg1, arg2, arg3);
}

/*
 *
 * Yield
 *
 */

void Yield(void)
{
  xTrace0(processswitch, TR_FULL_TRACE, "Yield");
  MASTER_UNLOCK;
  pthread_yield();
  MASTER_LOCK;
}

/*
 *
 * enter_master_sledgehammer_monitor
 *
 * Concurrency control is enforced in the xkernel by means of a monitor,
 *  in which all execution takes place. Here it is!
 *
 * A new thread begins here.
 * Inside the kernel:
 *      Dequeue a newThreadInfo block and do what it says
 * Outside the kernel:
 *      We are passed the argBlock.
 *
 * All the process block stuff looks totally defunct
 *
 */

static void *enter_master_sledgehammer_monitor(void *args) {
  CreateProcFunc	user_fn;
  int 			nargs;
  int 			threadcount;
  Process       	*pd;
  int 			*args_ptr = (int *)args;

	  /*  Acquire master kernel lock */
  xTrace0(processcreation, TR_MAJOR_EVENTS,
	  "enter_master_sledgehammer_monitor: entering\n");
  MASTER_LOCK;
  xTrace0(processcreation, TR_MAJOR_EVENTS,
	  "enter_master_sledgehammer_monitor: locked\n");

	  /*  Put a process descriptor on the list */
  pd = (Process *) xMalloc(sizeof(Process));
  pd -> link = NULL;
  pd -> thread = XTHREAD_SELF();

	  /*  Get function and arguments from array */
  user_fn = (CreateProcFunc) *(((int *)args)++);
  nargs = *(((int *)args)++);

  threadcount = current_xk_threads++;
  if (threadcount > max_xk_threads_inuse) {
    max_xk_threads_inuse = threadcount;
    xTrace1(processcreation, TR_EVENTS,
	    "processcreate: new thread high water mark %d", threadcount);
  }

  if (nargs > MAXARGS)
    Kabort("master_sledgehammer_monitor: cannot handle more than 6 args.");

  {
    int argv[MAXARGS];
    int i;
    
    for (i = 0; i < nargs; i++)
      argv[i] = *(((int *)args)++);
    xFree((char *) args_ptr);
    xTrace1(processcreation, TR_EVENTS,
	    "enter_master_sledgehammer_monitor: starting thread %d",
	    XTHREAD_SELF());

	    /*  Call routine with arguments */
    switch(nargs) {
      case 0: (user_fn)(0);
	break;
      case 1: (user_fn)(argv[0]);
	break;
      case 2: (user_fn)(argv[0], argv[1]);
	break;
      case 3: (user_fn)(argv[0], argv[1], argv[2]);
	break;
      case 4: (user_fn)(argv[0], argv[1], argv[2], argv[3]);
	break;
      case 5: (user_fn)(argv[0], argv[1], argv[2], argv[3], argv[4]);
	break;
      case 6: (user_fn)(argv[0], argv[1], argv[2], argv[3], argv[4],
			argv[5]);
      break;
    }
  }

  xFree((char *)pd);

  current_xk_threads--;
  MASTER_UNLOCK;
  xTrace1(processcreation, TR_MAJOR_EVENTS,
	  "enter_master_sledgehammer_monitor: thread %d unlocks",
	  XTHREAD_SELF());
  return NULL;
}


/*
 * 
 * Delay
 *        n in ms
 */
void Delay(int ms) {
  struct timeval	curtime;
  struct timespec	reqnap;	/* Requested nap time 			*/
  struct timespec	remnap;	/* Remaining nap time 			*/

#ifdef XK_DEBUG
  gettimeofday(&curtime, NULL);
  xTrace3(processswitch, TR_FULL_TRACE,
	  "Enter event function Delay(%d) @ %d:%06d",
	  ms, curtime.tv_sec, curtime.tv_usec);
#endif
  
  MASTER_UNLOCK;
  if ( ms ) {			/* Delay(0) same as Yield() 		*/
  
	    /* It is critical to correctly distribute the ms into the
	       nano and second portions. */
    reqnap.tv_sec  = (int)((float)ms/1000);
    reqnap.tv_nsec = (ms % 1000) * 1000000;
    while ( xnanosleep(&reqnap, &remnap) )
      reqnap = remnap;
  }
  MASTER_LOCK;
  
#ifdef XK_DEBUG  
  gettimeofday(&curtime, NULL);
  xTrace3(processswitch, TR_FULL_TRACE,
	  "Exit event function Delay(%d) @ %d:%06d",
	  ms, curtime.tv_sec, curtime.tv_usec);
#endif  
}


/*
 * signal_handler_thread
 * 
 *	Dedicated thread to sigwait on improtant signals other than those
 *	used internally by the x-kernel (e.g. SIGIO).
 */
static void *signal_handler_thread(void *arg) {
  sigset_t	set;

  sigemptyset(&set);
  sigaddset(&set,SIGINT);
  sigaddset(&set,SIGALRM);	/* catch spurious alarms from pthreads */
  while ( 1 ) {
    int	sig = 0;
    
    sigwait(&set, &sig);
    switch ( sig ) {
      case 0:
	xError("sigwait interrupted");
	break;
      case SIGINT:
	exit(0);
      case SIGALRM:
	xError("spurious SIGARLM caught");
	break;
      default:
	;
    }
  }
  Kabort("signal handling thread exiting");
}
