/*     
 * $RCSfile: total_order.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 *
 * $Log: total_order.c,v $
 * Revision 1.2  1994/05/06 20:33:58  menze
 *   [ 1994/05/06          menze ]
 *   Fixed declaration and string problems
 *   Use msgShow rather than MSG_SHOW
 *
 * Revision 1.1  1994/04/18  22:02:06  slm
 * Initial revision
 *
 */

#include "xkernel.h"
#include "psync.h"
#include "common.h"
#include "total_order.h"
#include "total_order_internal.h"
#include "utils.c"

#define MSG_SHOW	msgShow

       void           total_order_init(XObj);
static void           getproc_protl(XObj), getproc_sessn(XObj);
static XObj           total_open(XObj, XObj, XObj, Part *);
static xkern_return_t total_demux(XObj, XObj, Msg *);
static int            total_controlprotl(XObj, int, char *, int);
static xmsg_handle_t  total_push(XObj, Msg *);
static xkern_return_t total_pop(XObj, XObj, Msg *, VOID *);
static int            total_controlsessn(XObj, int, char *, int);
static long           rmsg_hdr_peek(VOID *, char *, long, VOID *);
static long           rmsg_hdr_strip(VOID *, char *, long, VOID *);
static void           smsg_hdr_store(VOID *, char *, long, VOID *);
static void           alive_timeout(Event, VOID *);
static void           update_wave(Sstate *, int, int);
static void           do_update(XObj, Sstate *);
static void           form_newcwave(Sstate *);
static int            complete_wave(Sstate *);
static void           sort(Sstate *);
static void           comp(IPhost *, int *, int, int);
static void           exchange(IPhost *, int *, int, int);
static void           remove_element(int *, int, int *);
static int            get_num(Sstate *);
static void           free_num(Sstate *, int);

int tracetotal_orderp = 1;

/*
 * Initialize the protocol state and the types of messages in which
 * TOTAL_ORDER is interested.
 */
void total_order_init(self)
XObj self;
{
    TOTALstate *pstate;

    xAssert(xIsProtocol(self));
    xTrace0(total_orderp, TR_GROSS_EVENTS, "TOTAL_ORDER init");
    if (!xIsProtocol(xGetDown(self,0)))
        Kabort("TOTAL_ORDER: could not get lower protocol DIVIDER");

    getproc_protl(self);

    if ((pstate = (TOTALstate *)xMalloc(sizeof(TOTALstate))) == NULL)
	Kabort("TOTAL_ORDER: malloc failed in total_order_init");
    self->state = (char *)pstate;
    pstate->map_demux  = mapCreate(100, sizeof(XObj));
    /* Indicate the types of msgs TOTAL_ORDER would like to see. */
    pstate->mytypes[0] = TOTAL_TYPE;
    pstate->mytypes[1] = MEMBERSHIP_TYPE;
    pstate->mytypes[2] = MONITOR_TYPE;
    pstate->mytypes[3] = RECOVERY_TYPE;
    pstate->mytypes[4] = TOTAL_ALIVE_TYPE;
    pstate->mytypes[5] = 0;
}

static void getproc_protl(self)
XObj self;
{
    xAssert(xIsProtocol(self));

    self->open    = total_open;
    self->demux   = total_demux;
    self->control = total_controlprotl;
}

static void getproc_sessn(self)
XObj self;
{
    xAssert(xIsSession(self));

    self->push    = total_push;
    self->pop     = total_pop;
    self->control = total_controlsessn;
}

static XObj total_open(self, hlpRcv, hlpType, p)
XObj self, hlpRcv, hlpType;
Part *p;	/* p = [psync1] */
/*
 * NOTE: Part list consists of the PSYNC session that was opened by DIVIDER.
 * This passing of an XObj in a part list is *** UNCONVENTIONAL *** and should
 * be avoided.
 */
{
    TOTALstate *pstate = (TOTALstate *)self->state;
    XObj       psync_sessn, divider_sessn, total_sessn;
    Part       part[MAX_TYPE+2];
    int        i;
    Sstate     *sstate;
    char       buf[sizeof(char) + MAX_PART*sizeof(PSYNCaddr)];

    xAssert(xIsProtocol(self));
    xTrace0(total_orderp, TR_FUNCTIONAL_TRACE,
	    "TOTAL_ORDER: entering total_open");

    psync_sessn = *(XObj *)partPop(p[0]);
    xTrace1(total_orderp, TR_DETAILED, "TOTAL_ORDER: psync_session = %d",
	    psync_sessn);

    /*
     * Prepare the part list for DIVIDER to be
     *   part = [psync_sessn,type1,...,typeN]
     * psync_sessn is a pointer to the PSYNC session which TOTAL_ORDER has to
     * set as its down[0] for optimization.
     */
    partInit(part, MAX_TYPE+2);
    partPush(part[0], (VOID *)&psync_sessn, sizeof(XObj));
    for (i = 0; pstate->mytypes[i]; i++)
        partPush(part[i+1], (VOID *)&pstate->mytypes[i], sizeof(int));
    partPush(part[i+1], (VOID *)NULL, 0);

    /* Open DIVIDER session. */
    divider_sessn = xOpen(self, self, xGetDown(self,0), part);

    /*
     * Create a TOTAL_ORDER session and set its down session to be the DIVIDER
     * session.
     */
    total_sessn = xCreateSessn(getproc_sessn, hlpRcv, hlpType, self, 1,
			       &divider_sessn);

    /* Bind the TOTAL_ORDER session to the DIVIDER session. */
    /* `map_demux' is <divider_sessn> --> <total_sessn> */
    mapBind(pstate->map_demux, (VOID *)&divider_sessn, (VOID *)total_sessn);

    if ((sstate = (Sstate *)xMalloc(sizeof(Sstate))) == NULL) {
	xError("TOTAL_ORDER: malloc failed in total_open");
	return ERR_XOBJ;
    }
    total_sessn->state = (char *)sstate;
    sstate->psync_sessn = psync_sessn;
    sstate->numcwave    = sstate->numlwave = 0;
    sstate->cwavep      = sstate->warr1;
    sstate->lwavep      = sstate->warr2;
    sstate->max_num     = sstate->nindex = 0;

    /*
     * Get the participant list from PSYNC.  This was given to the PSYNC
     * protocol when DIVIDER opens it as a response to COMPOSER's open.
     * *** UNCONVENTIONAL ***  TOTAL_ORDER should not directly contact PSYNC
     * since psync_sessn was obtained by passing it thru a part list.  Instead,
     * TOTAL_ORDER should do an xControl on DIVIDER and DIVIDER should pass it
     * on to PSYNC.
     */
    if (xControl(sstate->psync_sessn, PSYNC_GETPARTS, buf, sizeof(buf)) == -1) {
	xError("TOTAL_ORDER: PSYNC_GETPARTS returned error");
	return ERR_XOBJ;
    }
    sstate->numpart     = (int)buf[0];
    bcopy(buf+sizeof(char), (char *)sstate->parts, MAX_PART*sizeof(PSYNCaddr));
    sstate->alive_wait  = MAX_ALIVE_TIMEOUT;
    xTrace1(total_orderp, TR_DETAILED,
	    "TOTAL_ORDER: open scheduling alive_timeout for %d",
	    sstate->alive_wait);
    sstate->alive_event = evSchedule(alive_timeout, (VOID *)sstate,
				     sstate->alive_wait);
    sstate->map_table   = mapCreate(2*MAX_NODES, sizeof(char *));
    semInit(&sstate->Ordering, 1);
    xTrace0(total_orderp, TR_FUNCTIONAL_TRACE,
	    "TOTAL_ORDER: exiting total_open");
    return total_sessn;
}

static xkern_return_t total_demux(self, lls, msg)
XObj self, lls;
Msg *msg;
{
    XObj total_sessn;

    xTrace0(total_orderp, TR_FUNCTIONAL_TRACE,
	    "TOTAL_ORDER: entering total_demux");

    /* Pop the message to the appropriate TOTAL_ORDER session. */
    /* `map_demux' is <divider_sessn> --> <total_sessn> */
    if (mapResolve(((TOTALstate *)self->state)->map_demux, (VOID *)&lls,
		   (VOID **)&total_sessn) == XK_FAILURE)
        xError(
	    "TOTAL_ORDER: demux: can't find TOTAL_ORDER session; dropping msg");
    else
        xPop(total_sessn, lls, msg, (VOID *)NULL);

    xTrace0(total_orderp, TR_FUNCTIONAL_TRACE,
	    "TOTAL_ORDER: exiting total_demux");
    return XK_SUCCESS;
}

static int total_controlprotl(self, opcode, buf, len)
XObj self;
int opcode, len;
char *buf;
{
    xAssert(xIsProtocol(self));

    switch (opcode) {
        case TOTAL_RESTORE_SESSN:
            return 0;
        case TOTAL_GETCHECKPOINT:
            return 0;
        default:
            return xControl(xGetDown(self,0), opcode, buf, len);
    }
}

static xmsg_handle_t total_push(self, msg)
XObj self;
Msg *msg;
{
    Smsg_hdr      hdr;
    Sstate        *sstate = (Sstate *)self->state;
    xmsg_handle_t retval;

    xAssert(xIsSession(self));
    xTrace0(total_orderp, TR_FUNCTIONAL_TRACE,
	    "TOTAL_ORDER: entering total_push");
    xTrace1(total_orderp, TR_DETAILED, "TOTAL_ORDER: received msg of len %d",
            msgLen(msg));

    /* Put a header to the msg and push it to PSYNC.  hdr is [0,TOTAL_TYPE] */
    bzero((char *)&hdr, sizeof(Smsg_hdr));
#ifdef XK_FIRST_INT
    hdr.first = 0;
#else
    hdr.first = '0';
#endif
    hdr.type  = TOTAL_TYPE;
    xTrace2(total_orderp, TR_DETAILED,
            "TOTAL_ORDER: pushing hdr.first = %c, hdr.type = %d onto msg",
            hdr.first, hdr.type);
    msgPush(msg, smsg_hdr_store, (VOID *)&hdr, sizeof(Smsg_hdr), (VOID *)NULL);
    xTrace1(total_orderp, TR_DETAILED,
	    "TOTAL_ORDER: pushing msg of len %d", msgLen(msg));
    xIfTrace(total_orderp, TR_DETAILED) {
        MSG_SHOW(msg);
    }

    /* Cancel the alive event. */
    if (!evIsCancelled(sstate->alive_event))
        evCancel(sstate->alive_event);

    xTrace1(total_orderp, TR_DETAILED,
	    "TOTAL_ORDER: push: scheduling alive_timeout for %d",
	    sstate->alive_wait);
    sstate->alive_event = evSchedule(alive_timeout, (VOID *)sstate,
				     sstate->alive_wait);

    /* Push message directly to PSYNC.  *** UNCONVENTIONAL *** */
    retval = xPush(sstate->psync_sessn, msg);
    xTrace0(total_orderp, TR_FUNCTIONAL_TRACE,
	    "TOTAL_ORDER: exiting total_push");
    return retval;
}

static xkern_return_t total_pop(self, lls, msg, hdrp)
XObj self, lls;
Msg  *msg;
VOID *hdrp;
{
    Sstate    *sstate = (Sstate *)self->state;
    Rmsg_hdr  hdr;
    Strip_hdr dummyhdr;
    Msg       *msgcpy;
    int       num;

    xAssert(xIsSession(self));
    xTrace0(total_orderp, TR_FUNCTIONAL_TRACE,
	    "TOTAL_ORDER: entering total_pop");
    xTrace1(total_orderp, TR_DETAILED, "TOTAL_ORDER: received msg of len %d",
            msgLen(msg));

    semWait(&sstate->Ordering);
    /* Peek into the header from the msg.  hdr is [node,first,type,opcode] */
    msgPop(msg, rmsg_hdr_peek, (VOID *)&hdr, sizeof(Rmsg_hdr), (VOID *)NULL);
    xTrace4(total_orderp, TR_DETAILED,
     "TOTAL_ORDER: total_rmsg_hdr: node = %d first = %c, type = %s opcode = %d",
            hdr.node, hdr.first, type2str(hdr.type), hdr.opcode);

    /* Acknowledge the message. */
    if (xControl(sstate->psync_sessn, PSYNC_ACKMESSAGE, (char *)&hdr.node,
  		 sizeof(hdr.node)) == -1) {
  	xError("TOTAL_ORDER: PSYNC_ACKMESSAGE returned error");
  	return XK_FAILURE;
    }	

    /* Strip off the header from the msg (all but the opcode). */
    msgPop(msg, rmsg_hdr_strip, (VOID *)&dummyhdr, sizeof(Strip_hdr),
	   (VOID *)NULL);

    /* Make a copy of msg since it gets destroyed. */
    if ((msgcpy = (Msg *)xMalloc(sizeof(Msg))) == NULL) {
	xError("TOTAL_ORDER: malloc failed in total_pop");
	return XK_FAILURE;
    }
    msgConstructCopy(msgcpy, msg);

    /* Store the copy of the msg so it can be demuxed it when it is stable. */
    xTrace0(total_orderp, TR_DETAILED,
	    "TOTAL_ORDER: storing msg in membership array");
    xIfTrace(total_orderp, TR_DETAILED) {
        MSG_SHOW(msgcpy);
    }
    num = get_num(sstate);
    /* `map_table' is <node> --> <num> */
    mapBind(sstate->map_table, (VOID *)&hdr.node, (VOID *)num);
    xTrace3(total_orderp, TR_FUNCTIONAL_TRACE,
            "TOTAL_ORDER: binding node %d of type %s to num %d",
            hdr.node, type2str(hdr.type), num);
    sstate->membership[num].type   = hdr.type;
    sstate->membership[num].msg    = msgcpy;
    sstate->membership[num].mcwave = FALSE;
    update_wave(sstate, hdr.node, num);
    do_update(self, sstate);
    semSignal(&sstate->Ordering);

    xTrace0(total_orderp, TR_FUNCTIONAL_TRACE,
	    "TOTAL_ORDER: exiting total_pop");
    return XK_SUCCESS;
}

static int total_controlsessn(self, opcode, buf, len)
XObj self;
int  opcode, len;
char *buf;
{
    xAssert(xIsSession(self));

    return xControl(xGetDown(self,0), opcode, buf, len);
}

/*
 * Peek into the header of a msg.
 */
static long rmsg_hdr_peek(hdr, src, len, arg)
VOID *hdr, *arg;
char *src;
long len;
{
    bcopy(src, (char *)hdr, len);
    return 0;
}

/*
 * Strip off the header of a msg.
 */
static long rmsg_hdr_strip(hdr, src, len, arg)
VOID *hdr, *arg;
char *src;
long len;
{
    bcopy(src, (char *)hdr, len);
    return len;
}

/*
 * Add a header of type Smsg_hdr to the msg.
 */
static void smsg_hdr_store(hdr, dst, len, arg)
VOID *hdr, *arg;
char *dst;
long len;
{
    bcopy((char *)hdr, dst, len);
}

/*
 * Fire up every few ms and send an ALIVE ack to the remote host thru PSYNC's
 * broadcast.
 */
static void alive_timeout(ev, arg)
Event ev;
VOID  *arg;
{
    Sstate    *sstate = (Sstate *)arg;
    Alive_hdr hdr;
    Msg       msg;

    /* Construct a dummy msg and send it. */
    evDetach(sstate->alive_event);
    bzero((char *)&hdr, sizeof(Alive_hdr));
#ifdef XK_FIRST_INT
    hdr.first = 0;
#else
    hdr.first = '0';
#endif
    hdr.type   = TOTAL_ALIVE_TYPE;
    hdr.opcode = 0;		/* dummy field */
    msgConstructBuffer(&msg, (char *)&hdr, sizeof(Alive_hdr));

    /* Pushing directly to PSYNC.  *** UNCONVENTIONAL *** */
    xTrace0(total_orderp, TR_DETAILED,
     "TOTAL_ORDER: alive_timeout: pushing message to PSYNC (UNCONVENTIONAL!?)");
    xPush(sstate->psync_sessn, &msg);
    msgDestroy(&msg);
    xTrace1(total_orderp, TR_DETAILED,
	    "TOTAL_ORDER: alive_timeout: rescheduling itself for %d later",
  	    sstate->alive_wait);
    sstate->alive_event = evSchedule(alive_timeout, (VOID *)sstate,
				     sstate->alive_wait);
}

/*
 * For an explanation about the algorithm and routines used to impose total
 * ordering on the set of messages, refer to shiv's dissertation.
 */

static void update_wave(sstate, node, num)
Sstate *sstate;
int    node, num;
{
    int  i;
    char buf[2*sizeof(int)];

    for (i = 0; i < sstate->numcwave; i++) {
        bcopy((char *)&sstate->cwavep[i], buf, sizeof(int));
        bcopy((char *)&node, buf+sizeof(int), sizeof(int));
        if (xControl(sstate->psync_sessn, PSYNC_PRECEDES, buf, sizeof(buf)) ==
	        -1)
	    Kabort("TOTAL_ORDER: PSYNC_PRECEDES returned error");
	if ((int)buf[0] == TRUE)	/* a node in the current wave */
	    return;			/*   precedes this node       */
        /*
         * Note: any nodes that are not put in current wave at this time should
	 * be found later as next nodes of a node currently in current wave
	 * (after it has become last wave) (see form_newcwave).
         */
    }
    /* Add node to current wave. */
    sstate->cwavep[sstate->numcwave++] = node;
    sstate->membership[num].mcwave     = TRUE;
    if (sstate->numcwave > MAX_NODES)
	Kabort("TOTAL_ORDER: too many nodes in sstate cwave array");
}

static void do_update(self, sstate)
XObj   self;
Sstate *sstate;
{
    int num, i;

    while (complete_wave(sstate)) {	/* current wave is complete */
        sort(sstate);
        for (i = 0; i < sstate->numcwave; i++) {
            xTrace1(total_orderp, TR_DETAILED,
	            "TOTAL_ORDER: do_update: trying to resolve node %d",
		    sstate->cwavep[i]);
            if (mapResolve(sstate->map_table, (VOID *)&sstate->cwavep[i],
		           (VOID **)&num) == XK_FAILURE)
	        Kabort("TOTAL_ORDER: mapResolve failed in do_update");
            xTrace1(total_orderp, TR_DETAILED,
	            "TOTAL_ORDER:            node resolved to num %d", num);
            xTrace1(total_orderp, TR_FUNCTIONAL_TRACE,
                    "TOTAL_ORDER:            apply opcode if %s == TOTAL",
                    type2str(sstate->membership[num].type));
            if (sstate->membership[num].type == TOTAL_TYPE) {
                xTrace1(total_orderp, TR_DETAILED,
                        "TOTAL_ORDER:            demuxing msg of len %d",
                        msgLen(sstate->membership[num].msg));
                xIfTrace(total_orderp, TR_DETAILED) {
                    MSG_SHOW(sstate->membership[num].msg);
                }
		/* Send the message up. */
                xDemux(self, sstate->membership[num].msg);
            }
        }
        form_newcwave(sstate);
    }
}

/*
 * Form new "current wave" of nodes.
 */
static void form_newcwave(sstate)
Sstate *sstate;
{
    int  i, j, k, num, num2, *tmpptr, *node, num_next, not_inwave;
    char buf1[(MAX_PART+1)*sizeof(int)], buf2[2*sizeof(int)];

    xTrace0(total_orderp, TR_FUNCTIONAL_TRACE,
	    "TOTAL_ORDER: entering form_newcwave");

    /* For all nodes in current wave, turn off `member of current wave' flag. */
    for (i = 0; i < sstate->numcwave; i++) {
        xTrace1(total_orderp, TR_DETAILED,
		"TOTAL_ORDER: trying to resolve node %d", sstate->cwavep[i]);
        if (mapResolve(sstate->map_table, (VOID *)&sstate->cwavep[i],
		       (VOID **)&num) == XK_FAILURE)
	    Kabort(
	"TOTAL_ORDER: mapResolve failed in form_newcwave for node in current wave");
        xTrace1(total_orderp, TR_DETAILED,
	        "TOTAL_ORDER: node resolved to num %d", num);
        sstate->membership[num].mcwave = FALSE;
    }

    /* Unbind all nodes in last wave. */
    for (i = 0; i < sstate->numlwave; i++) {
        xTrace1(total_orderp, TR_DETAILED,
	        "TOTAL_ORDER: trying to resolve node %d", sstate->lwavep[i]);
        if (mapResolve(sstate->map_table, (VOID *)&sstate->lwavep[i],
		       (VOID **)&num) == XK_FAILURE)
	    Kabort(
       "TOTAL_ORDER: mapResolve failed in form_newcwave for node in last wave");
        xTrace1(total_orderp, TR_DETAILED,
                "TOTAL_ORDER: node resolved to num %d; unbind node", num);
        mapUnbind(sstate->map_table, (VOID *)&sstate->lwavep[i]);
        msgDestroy(sstate->membership[num].msg);
        xFree((char *)sstate->membership[num].msg);
        remove_element(sstate->lwavep, sstate->lwavep[i], &sstate->numlwave);
        i--;
	free_num(sstate, num);
    }

    /* Change current wave to last wave. */
    tmpptr           = sstate->lwavep;
    sstate->lwavep   = sstate->cwavep;
    sstate->numlwave = sstate->numcwave;
    sstate->cwavep   = tmpptr;
    sstate->numcwave = 0;
    /*
     * For each node in last wave (previously current wave), get all its
     * next nodes and determine if they should go in the new current wave.
     */
    for (i = 0; i < sstate->numlwave; i++) {
        bcopy((char *)&sstate->lwavep[i], buf1, sizeof(int));
        if (xControl(sstate->psync_sessn, PSYNC_NEXTNODES, buf1,
		     sizeof(buf1)) == -1)
	    Kabort("TOTAL_ORDER: PSYNC_NEXTNODES returned error");
	bcopy(buf1, (char *)&num_next, sizeof(int));
	node = (int *)(buf1 + sizeof(int));
        for (j = 0; j < num_next; j++) {
            xTrace1(total_orderp, TR_DETAILED,
	            "TOTAL_ORDER: trying to resolve node %d", node[j]);
	    if (mapResolve(sstate->map_table, (VOID *)&node[j],
			   (VOID **)&num) == XK_FAILURE) {
                xError(
	    "TOTAL_ORDER: mapResolve failed in form_newcwave for a next node");
                Kabort(
	    "             of a last wave node; msg should have been received");
	    }
            xTrace1(total_orderp, TR_DETAILED,
	            "TOTAL_ORDER: node resolved to num %d", num);
            if (sstate->membership[num].mcwave == FALSE) {
                not_inwave = TRUE;
                for (k = 0; k < sstate->numcwave; k++) {
                    bcopy((char *)&sstate->cwavep[k], buf2, sizeof(int));
                    bcopy((char *)&node[j], buf2+sizeof(int), sizeof(int));
                    if (xControl(sstate->psync_sessn, PSYNC_PRECEDES, buf2,
				 sizeof(buf2)) == -1)
	                Kabort("TOTAL_ORDER: PSYNC_PRECEDES returned error");
		    if ((int)buf2[0] == TRUE) {	/* a node in cwave precedes */
                        not_inwave = FALSE;	/*   this node; this node   */
		    }				/*   will not go into cwave */
		    else {	/* this node precedes a node in cwave;   */
				/*   remove the preceded node from cwave */
                        remove_element(sstate->cwavep, sstate->cwavep[k],
  				       &sstate->numcwave);
                        xTrace1(total_orderp, TR_DETAILED,
	                        "TOTAL_ORDER: trying to resolve node %d",
			        sstate->cwavep[k]);
			if (mapResolve(sstate->map_table,
				       (VOID *)&sstate->cwavep[k],
				       (VOID **)&num2) == XK_FAILURE) {
                            xError(
		  "TOTAL_ORDER: mapResolve failed in form_newcwave for a node");
                            Kabort(
		  "             in current wave that needs to be removed");
			}
                        xTrace1(total_orderp, TR_DETAILED,
	                        "TOTAL_ORDER: node resolved to num %d", num2);
			sstate->membership[num2].mcwave = FALSE;
                        k--;
                    }
                }
                if (not_inwave) {
                    sstate->membership[num].mcwave     = TRUE;
                    sstate->cwavep[sstate->numcwave++] = node[j];
                    if (sstate->numcwave > MAX_NODES)
	                Kabort(
		           "TOTAL_ORDER: too many nodes in sstate cwave array");
                }
            }
        }
    }
}

/*
 * Return TRUE if any node in current wave is stable.
 */
static int complete_wave(sstate)
Sstate *sstate;
{
    int  i;
    char buf[sizeof(int)];

    for (i = 0; i < sstate->numcwave; i++) {
	bcopy((char *)&sstate->cwavep[i], buf, sizeof(int));
        if (xControl(sstate->psync_sessn, PSYNC_STABLE, buf, sizeof(int)) == -1)
	    Kabort("TOTAL_ORDER: PSYNC_STABLE returned error");
	if ((int)buf[0] == TRUE) {
	    xTrace0(total_orderp, TR_DETAILED,
		    "TOTAL_ORDER: PSYNC_STABLE returned TRUE");
	    return TRUE;
        }
    }
    xTrace0(total_orderp, TR_DETAILED,
	    "TOTAL_ORDER: PSYNC_STABLE returned FALSE");
    return FALSE;
}

/*
 * Sort nodes in current wave by IP address of sender.
 */
static void sort(sstate)
Sstate *sstate;
{
    int       i, j;
    IPhost    address[MAX_PART];
    char      buf[MAX(sizeof(int), sizeof(PSYNCaddr))];
    PSYNCaddr *ps;

    bzero((char *)address, MAX_PART * sizeof(IPhost));
    for (i = 1; i < sstate->numcwave; i++) {
        for (j = 0; j < i; j++) {
            if (address[j].a == 0) {
                bcopy((char *)&sstate->cwavep[j], buf, sizeof(int));
                if (xControl(sstate->psync_sessn, PSYNC_SENDER, buf,
			     sizeof(buf)) == -1)
		    Kabort("TOTAL_ORDER: PSYNC_SENDER returned error");
                ps = (PSYNCaddr *)buf;
                address[j] = ps->host;
            }
            if (address[j+1].a == 0) {
                bcopy((char *)&sstate->cwavep[j+1], buf, sizeof(int));
                if (xControl(sstate->psync_sessn, PSYNC_SENDER, buf,
			     sizeof(buf)) == -1)
		    Kabort("TOTAL_ORDER: PSYNC_SENDER returned error");
                ps = (PSYNCaddr *)buf;
                address[j+1] = ps->host;
            }
            comp(address, sstate->cwavep, j, j+1);
        }
    }
}

/*
 * Compare IP addresses.
 */
static void comp(address, wave, i, j)
IPhost *address;
int    *wave, i, j;
{
    if (address[i].a > address[j].a)
        exchange(address, wave, i, j);
    else if (address[i].a == address[j].a) {
        if (address[i].b > address[j].b)
            exchange(address, wave, i, j);
        else if (address[i].b == address[j].b) {
            if (address[i].c > address[j].c)
                exchange(address, wave, i, j);
            else if (address[i].c == address[j].c) {
                if (address[i].d > address[j].d)
                    exchange(address, wave, i, j);
	    }
	}
    }
}

static void exchange(address, wave, i, j)
IPhost *address;
int    *wave, i, j;
{
    IPhost tmpaddr;
    int    tmpnode;

    tmpaddr    = address[i];
    tmpnode    = wave[i];
    wave[i]    = wave[j];
    address[i] = address[j];
    wave[j]    = tmpnode;
    address[j] = tmpaddr;
}

static void remove_element(warray, element, size)
int *warray, element, *size;
{
    int i;

    for (i = 0; i < *size; i++)
        if (warray[i] == element) {
            (*size)--;
            if (*size > 0)
		warray[i] = warray[*size];
            return;
        }
}

static int get_num(sstate)
Sstate *sstate;
{
    int num;

    if (sstate->nindex > 0) {
	sstate->nindex--;
	if (sstate->num_arr[sstate->nindex] == sstate->max_num)
	    sstate->max_num++;
	num = sstate->num_arr[sstate->nindex];
    }
    else {
        sstate->max_num++;
        num = sstate->max_num - 1;
    }
    if (num >= MAX_NODES)
        Kabort("TOTAL_ORDER: too many nodes for membership array");
    return num;
}

static void free_num(sstate, num)
Sstate *sstate;
int    num;
{
    if (num == sstate->max_num - 1)
	sstate->max_num--;
    else
	sstate->num_arr[sstate->nindex++] = num;
}
