/*
 * tcp_fsm.h
 *
 * Derived from:
 *
 * Copyright (c) 1982, 1986 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of California at Berkeley. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 *
 *	@(#)tcp_fsm.h	7.2 (Berkeley) 12/7/87
 *
 * Modified for x-kernel v3.2
 * Modifications Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 * $Revision: 1.1 $
 * $Date: 1996/07/11 21:39:21 $
 */

/*
 * TCP FSM state definitions.
 * Per RFC793, September, 1981.
 */

#define	TCP_S_NSTATES	11

#define	TCP_SS_CLOSED		0	/* closed */
#define	TCP_SS_LISTEN		1	/* listening for connection */
#define	TCP_SS_SYN_SENT		2	/* active, have sent syn */
#define	TCP_SS_SYN_RECEIVED	3	/* have send and received syn */
/* states < TCP_SS_ESTABLISHED are those where connections not established */
#define	TCP_SS_ESTABLISHED	4	/* established */
#define	TCP_SS_CLOSE_WAIT		5	/* rcvd fin, waiting for close */
/* states > TCP_SS_CLOSE_WAIT are those where user has closed */
#define	TCP_SS_FIN_WAIT_1		6	/* have closed, sent fin */
#define	TCP_SS_CLOSING		7	/* closed xchd FIN; await FIN ACK */
#define	TCP_SS_LAST_ACK		8	/* had fin and close; await FIN ACK */
/* states > TCP_SS_CLOSE_WAIT && < TCP_SS_FIN_WAIT_2 await ACK of FIN */
#define	TCP_SS_FIN_WAIT_2		9	/* have closed, fin is acked */
#define	TCP_SS_TIME_WAIT		10	/* in 2*msl quiet wait after close */

#define	TCP_SS_HAVERCVDSYN(s)	((s) >= TCP_SS_SYN_RECEIVED)
#define	TCP_SS_HAVERCVDFIN(s)	((s) >= TCP_SS_TIME_WAIT)

#ifdef	TCP_SOUTFLAGS
/*
 * Flags used when sending segments in tcp_s_output.
 * Basic flags (TH_RST,TH_ACK,TH_SYN,TH_FIN) are totally
 * determined by state, with the proviso that TH_FIN is sent only
 * if all data queued for output is included in the segment.
 */
u_char	tcp_s_outflags[TCP_S_NSTATES] = {
    TH_RST|TH_ACK, 0, TH_SYN, TH_SYN|TH_ACK,
    TH_ACK, TH_ACK,
    TH_FIN|TH_ACK, TH_FIN|TH_ACK, TH_FIN|TH_ACK, TH_ACK, TH_ACK,
};
#endif

#ifdef KPROF
int	tcp_s_acounts[TCP_S_NSTATES][PRU_NREQ];
#endif

#ifdef	TCP_SSTATES
char *tcp_sstates[] = {
	"CLOSED",	"LISTEN",	"SYN_SENT",	"SYN_RCVD",
	"ESTABLISHED",	"CLOSE_WAIT",	"FIN_WAIT_1",	"CLOSING",
	"LAST_ACK",	"FIN_WAIT_2",	"TIME_WAIT",
};
#endif
