/*
 * $RCSfile: tcp_timer.h,v $
 *
 * Derived from:
 *
 * Copyright (c) 1982, 1986 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of California at Berkeley. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 *
 *	@(#)tcp_timer.h	7.4 (Berkeley) 12/7/87
 *
 * Modified for x-kernel v3.2
 * Modifications Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 * $Log: tcp_timer.h,v $
 * Revision 1.1  1996/07/11 21:39:21  nicko
 * Initial revision
 *
 * Revision 1.3  1993/12/16  01:44:39  menze
 * *** empty log message ***
 *
 */

/*
 * Definitions of the TCP_S timers.  These timers are counted
 * down PR_SLOWHZ times a second.
 */
#define	TCP_ST_NTIMERS	4

#define	TCP_ST_REXMT	0		/* retransmit */
#define	TCP_ST_PERSIST	1		/* retransmit persistance */
#define	TCP_ST_KEEP	2		/* keep alive */
#define	TCP_ST_2MSL	3		/* 2*msl quiet time timer */

/*
 * The TCP_ST_REXMT timer is used to force retransmissions.
 * The TCP_S has the TCP_ST_REXMT timer set whenever segments
 * have been sent for which ACKs are expected but not yet
 * received.  If an ACK is received which advances tp->snd_una,
 * then the retransmit timer is cleared (if there are no more
 * outstanding segments) or reset to the base value (if there
 * are more ACKs expected).  Whenever the retransmit timer goes off,
 * we retransmit one unacknowledged segment, and do a backoff
 * on the retransmit timer.
 *
 * The TCP_ST_PERSIST timer is used to keep window size information
 * flowing even if the window goes shut.  If all previous transmissions
 * have been acknowledged (so that there are no retransmissions in progress),
 * and the window is too small to bother sending anything, then we start
 * the TCP_ST_PERSIST timer.  When it expires, if the window is nonzero,
 * we go to transmit state.  Otherwise, at intervals send a single byte
 * into the peer's window to force him to update our window information.
 * We do this at most as often as TCP_ST_PERSMIN time intervals,
 * but no more frequently than the current estimate of round-trip
 * packet time.  The TCP_ST_PERSIST timer is cleared whenever we receive
 * a window update from the peer.
 *
 * The TCP_ST_KEEP timer is used to keep connections alive.  If an
 * connection is idle (no segments received) for TCP_STV_KEEP amount of time,
 * but not yet established, then we drop the connection.  If the connection
 * is established, then we force the peer to send us a segment by sending:
 *	<SEQ=SND.UNA-1><ACK=RCV.NXT><CTL=ACK>
 * This segment is (deliberately) outside the window, and should elicit
 * an ack segment in response from the peer.  If, despite the TCP_ST_KEEP
 * initiated segments we cannot elicit a response from a peer in TCP_ST_MAXIDLE
 * amount of time, then we drop the connection.
 */

#define	TCP_S_TTL		30		/* default time to live for TCP_S segs */
extern int	tcp_s_ttl;		/* time to live for TCP_S segs */
/*
 * Time constants.
 */
#define	TCP_STV_MSL	( 15*PR_SLOWHZ)		/* max seg lifetime */
#define	TCP_STV_SRTTBASE	0			/* base roundtrip time;
						   if 0, no idea yet */
#define	TCP_STV_SRTTDFLT	(  3*PR_SLOWHZ)		/* assumed RTT if no info */

#define	TCP_STV_PERSMIN	(  5*PR_SLOWHZ)		/* retransmit persistance */
#define	TCP_STV_PERSMAX	( 60*PR_SLOWHZ)		/* maximum persist interval */

#define	TCP_STV_KEEP	( 75*PR_SLOWHZ)		/* keep alive - 75 secs */
#define	TCP_STV_MAXIDLE	(  8*TCP_STV_KEEP)	/* maximum allowable idle
						   time before drop conn */

#define	TCP_STV_MIN	(  1*PR_SLOWHZ)		/* minimum allowable value */
#define	TCP_STV_REXMTMAX	( 64*PR_SLOWHZ)		/* max allowable REXMT value */

#define	TCP_S_LINGERTIME	120			/* linger at most 2 minutes */

#define	TCP_S_MAXRXTSHIFT	12			/* maximum retransmits */

#ifdef	TCP_STIMERS
char *tcp_stimers[] =
    { "REXMT", "PERSIST", "KEEP", "2MSL" };
#endif

/*
 * Force a time value to be in a certain range.
 */
#define	TCP_ST_RANGESET(tv, value, tvmin, tvmax) { \
	(tv) = (value); \
	if ((tv) < (tvmin)) \
		(tv) = (tvmin); \
	else if ((tv) > (tvmax)) \
		(tv) = (tvmax); \
}


extern int tcp_s_backoff[];


/*
 * Intervals are in microseconds
 */
#define TCP_S_FAST_INTERVAL 200 * 1000
#define TCP_S_SLOW_INTERVAL 500 * 1000	
