/*
 * $RCSfile: mem.c,v $
 *
 * Universal Stub Compiler
 *
 * Copyright (c) 1994    Arizona Board of Regents
 *
 * $Revision: 1.1 $
 * $Date: 1994/10/13 17:43:30 $
 */

#include "usc.h"
#include "mem.h"
#include "debug.h"

typedef struct garbage {
  void *ptr;
  size_t size;
  struct garbage *next;
} Garbage;

Garbage *the_garbage;
int total_garbage;
int total_malloced;

void add_ptr (void *ptr, size_t size)
{
  Garbage *temp = (Garbage *)Malloc (sizeof (Garbage));
#ifdef GARBAGE_CHECK
  Garbage *p;
  for (p = the_garbage; p; p = p->next)
    if (p->ptr == ptr)
      crash ("Duplicate pointer in garbage! Ptr == %x", ptr);
#endif
  temp->ptr = ptr;
  temp->size = size;
  temp->next = the_garbage;
  the_garbage = temp;
}

void AddToGarbage (void *ptr, size_t size)
{
  /*
  dprintf (260, "addtogarbage: %d", size);
  */
  dprintf (860, "addtogarbage: %d", size);
  add_ptr (ptr, size);
}
  

void Free (void *ptr, size_t size)
{
  if (ptr) {
    dprintf (1000, "freeing: 0x%x %d", ptr, size);

#ifdef USE_DMALLOC
    dfree (ptr);
#else
    free (ptr);
#endif
    total_malloced -= (int)size;
    dprintf (1000, "Total Malloc = %d\n", total_malloced);
  }
}

void *Malloc (size_t size)
{
  void *temp;
/*   assert (size >= 0); */
  dprintf (1000, "entering Malloc");
  dprintf (860, "Malloc: %d", size);
  /*
  dprintf (260, "Malloc: %d", size);
  */
#ifdef USE_DMALLOC
  temp = dmalloc (size);
#else
  temp = malloc (size);
#endif
  if (!temp)
    crash ("malloc error\n");
  total_malloced += (int)size;
  dprintf (1000, "exiting Malloc");
  return temp;
}

void *TempMalloc (size_t size)
{
  void *temp;
  dprintf (1000, "entering TempMalloc");
  dprintf (860, "TempMalloc: %d", size);
  /*
  dprintf (260, "TempMalloc: %d", size);
  */
#ifdef USE_DMALLOC
  temp = dmalloc (size);
#else
  temp = malloc (size);
#endif
  total_malloced += size;
  if (!temp)
    crash ("malloc error\n");
  add_ptr (temp, size);
  dprintf (1000, "exiting TempMalloc");
  return temp;
}

void CleanGarbage (void)
{
  Garbage *temp = the_garbage;
  while (the_garbage) {
    the_garbage = the_garbage->next;
#ifdef USE_DMALLOC
    dfree (temp->ptr);
#else
    free (temp->ptr);
#endif
    total_garbage += temp->size;
    total_malloced -= temp->size;
    /*
    dprintf (260, "Nuking: %d", temp->size);
    */
    dprintf (860, "Nuking: %d", temp->size);
    Free (temp, sizeof (Garbage));
    temp = the_garbage;
  }
}

void FreeGarbage (void *ptr)
{
}

void init_mem (void)
{
  the_garbage = NULL;
  total_garbage = 0;
  total_malloced = 0;
}

#if 0

void print_mallinfo (int level)
{
  struct mallinfo mal_info;
  if (level <= DebugLevel) {
    mal_info = mallinfo();
    printf ("Mallinfo:\n");
    printf ("arena size: %d\n", mal_info.arena);
    printf ("ordblks: %s\n", mal_info.ordblks);
    printf ("smblks: %d\n", mal_info.smblks);
    printf ("");
  }
}

#endif

void close_mem (void)
{
  dprintf (710 /*110*/, "Total garbage collected: %d\n", total_garbage);
  dprintf (710 /*110*/, "Total left malloced: %d\n", total_malloced);
}
