%{
/*
 * $RCSfile: types.lex,v $
 *
 * Universal Stub Compiler
 *
 * Copyright (c) 1994    Arizona Board of Regents
 *
 * $Log: types.lex,v $
 * Revision 1.3  1996/02/15 22:04:04  mjk
 * Works with flex.
 * Define LEXTRACE to enable TraceRule() macro for debugging.
 * Counts line correctly now (the final section never counted lines)
 *
 * Revision 1.2  1994/10/13  17:49:42  dorgival
 * Protected access to yydebug in yyinit with "#ifdef YYDEBUG"
 *
 * Revision 1.1  1994/10/13  17:47:41  dorgival
 * Initial revision
 *
 */

/* Lexical analyzer for C: Contains some lex specific constructs to *
 * handle exclusive initial states (for the %% delimiters), does not
 * work with flex as it is - should be fixed later
 *
 * THE ABOVE COMMENT IS NO LONGER TRUE.  To see this define LEXTRACE
 * and compare output of a lex and flex based scanner.  */

#include "usc.h"
#include "symbol.h"
#include "debug.h"
#include "buffer.h"
#include "y.tab.h"

#ifdef NO_BSEARCH
#define bsearch(key,base,nel,size,compar) dumb_search((key)->name)
#else
#include <search.h>
#endif

				/* Define LEXTRACE to enable rule
                                   based tracing. */
/*
#define LEXTRACE
*/


#ifdef LEXTRACE  
# define TraceRule(a)	fprintf(stderr, "%s: %d[%s]\n", a, LineNum, yytext)
#else
# define TraceRule(a)
#endif  

extern int yydebug;
int LineNum = 1;
Buffer *SourceText;

#ifdef FLEX_SCANNER
extern char *yytext;
#endif

void add_text (void)
{
  writebuf (SourceText, yytext);
}

int id_or_keyword_or_type (char *lex);

%}

let	[_a-zA-Z]
alnum	[_a-zA-Z0-9]
h	[0-9a-fA-F]
o	[0-7]
d	[0-9]
suffix	[UuLl]
white	[ \t\f]
newline [\n\r]

%s HEADER TRAILER

%p 3000

%{
  static int first_time = 1;
%}

%%

%{
  if ( first_time ) {
    BEGIN HEADER;
    first_time = 0;
  }
%}

<HEADER>^%%.*$         {
                         TraceRule("<HEADER>^%%.*$");
			 if ( do_preprocess )
			   fprintf(OutFile,"%s\n",yytext);
			 BEGIN 0;
			}
<HEADER>^[^%\n].*$      {
                          TraceRule("<HEADER>^[^%\\n].*$");
			  fprintf(OutFile,"%s\n",yytext);
                        }
<HEADER>^%[^%\n].*$     {
                          TraceRule("<HEADER>^%[^%\\n].*$");
                          fprintf(OutFile,"%s\n",yytext);
                        }
<HEADER>^%$             {
                          TraceRule("<HEADER>^%$");
                          fprintf(OutFile,"%s\n",yytext);
                        }
<HEADER>newline         {
                          TraceRule("<HEADER>newline");
                          fprintf(OutFile,"\n");
                        }

^"%%".*$                {
			  extern void finishup(void);

			  TraceRule("^\"%%\".*$");

			  finishup();

			  if ( do_preprocess )
			    fprintf(OutFile,"%s\n",yytext);
			  else
			    fprintf(OutFile,"\n");

			  BEGIN TRAILER;
			}


<TRAILER>\n             {
                          TraceRule("<TRAILER>\\n");
			  LineNum++;
                          fprintf(OutFile,"%s",yytext);
                        }
<TRAILER>^.*$           {
                          TraceRule("<TRAILER>^.*$");
                          fprintf(OutFile,"%s",yytext);
                        }

"/*"  {	 
  int i, lasti = 0;

  TraceRule("\"/*\" - begin comment");
  
  dprintf (1000, "COMMENT");
  
				/* Digest the rest of the comment */
  while ( (i = input()) && i != EOF) {
    if (i == '\n' || i == '\r')
      LineNum++;
    if (i == '/' && lasti == '*')
      break;
    lasti = i;
  }
				/* Trace after LineNum has been
                                   corrected. */
  TraceRule("\"/*\" - end comment");

  if (i == 0 || i == EOF)
    yyerror ("End of file in comment.\n");
}

\"(\\.|[^\"])*\"    {
  TraceRule("\\\"(\\\\.|[^\\\"]]*\\\"");
  dprintf (1000, "STRING"); add_text(); 
  set_to_string (yylval.theconst, yytext); 
  return STRING; 
}

\"(\\.|[^\"])*[\r\n] {
  TraceRule("\\\"(\\\\.|[^\\\"]]*[\\r\\n]");
  yyerror ("Adding missing \" to string constant\n");
  add_text(); 
  set_to_string (yylval.theconst, yytext); return STRING;
}

'.'   	       	     |
'\\.' 	       	     |
'\\{o}({o}{o}?)?'    |
'\\x{h}({h}{h}?)?'   |
0{o}*{suffix}? 	     |
0x{h}+{suffix}?	     |
[1-9]{d}*{suffix}?    {
  TraceRule("ICON");
  dprintf (1000, "ICON");
  add_text(); 
  set_to_int (yylval.theconst, yytext);
  return ICON;
}

({d}+|{d}+\.[^.]{d}*|{d}*\.{d}+)([eE][-+]?{d}+)?[fF]? {
  TraceRule("FCON");
  dprintf (1000, "FCON");
  add_text(); 
  set_to_float (yylval.theconst, yytext);
  return FCON;
}

"("   	 { dprintf (1000, "LP"); add_text(); return yytext[0]; }
")"   	 { dprintf (1000, "RP"); add_text(); return yytext[0]; }
"{"   	 { dprintf (1000, "LC"); pushlevel(); add_text(); return yytext[0]; }
"}"   	 { dprintf (1000, "RC"); poplevel(); add_text(); return yytext[0]; }
"["   	 { dprintf (1000, "LB"); add_text(); return yytext[0]; }
"]"   	 { dprintf (1000, "RB"); add_text(); return yytext[0]; }
"#"   	 { dprintf (1000, "HASH"); add_text(); return HASH; }

"->"      { yylval.theop = ARROW; add_text(); return ARROW; }
"..."     { dprintf (1000, "DOTDOTDOT"); add_text(); return DOTDOTDOT;     }
".."      { dprintf (1000, "DOTDOT"); add_text(); return DOTDOT;     }
"."       { yylval.theop = DOT; add_text(); return DOT;     }

"++"  	 { yylval.theop = PLUSPLUS; add_text(); return PLUSPLUS;}
"--"  	 { yylval.theop = MINMIN; add_text(); return MINMIN; }

"~"		 { yylval.theop = TILDE; add_text(); return yytext[0]; }
"!"  	 	 { yylval.theop = BANG; add_text(); return yytext[0]; }

"*"   	 { yylval.theop = STAR; add_text(); return yytext[0]; }
"/"		 { yylval.theop = DIV; add_text(); return yytext[0]; }
"%" 		 { yylval.theop = MOD; add_text(); return yytext[0]; }
"+"   	 { yylval.theop = PLUS; add_text(); return yytext[0]; }
"-"   	 { yylval.theop = MINUS; add_text(); return yytext[0]; }
"<<"		 { yylval.theop = LSHIFT; add_text(); return LSHIFT; }
">>"  	 { yylval.theop = RSHIFT; add_text(); return RSHIFT; }
">="		 { yylval.theop = GE; add_text(); return GE; }
"<="	 	 { yylval.theop = LE; add_text(); return LE; }
"<"		 { yylval.theop = LT; add_text(); return LT; }
">"		 { yylval.theop = GT; add_text(); return GT; }
"!="		 { yylval.theop = NEQ; add_text(); return NEQ; }
"=="		 { yylval.theop = EQEQ; add_text(); return EQEQ; }

"+="     { dprintf (1000, "ADDEQ"); add_text(); return ADDEQ; }
"-="     { dprintf (1000, "SUBEQ"); add_text(); return SUBEQ; }
"*="     { dprintf (1000, "MULEQ"); add_text(); return MULEQ; }
"/="     { dprintf (1000, "DIVEQ"); add_text(); return DIVEQ; }
"%="     { dprintf (1000, "MODEQ"); add_text(); return MODEQ; }
"&="     { dprintf (1000, "ANDEQ"); add_text(); return ANDEQ; }
"|="     { dprintf (1000, "OREQ"); add_text(); return OREQ; }
"^="     { dprintf (1000, "XOREQ"); add_text(); return XOREQ; }
">>="     { dprintf (1000, "RSHIFTEQ"); add_text(); return RSHIFTEQ; }
"<<="     { dprintf (1000, "LSHIFTEQ"); add_text(); return LSHIFTEQ; }

"="   	 { yylval.theop = EQUAL; add_text(); return EQUAL; }
"&"   	 { yylval.theop = AND; add_text(); return AND; }
"^"   	 { yylval.theop = XOR; add_text(); return XOR; }
"|"   	 { yylval.theop = OR; add_text(); return OR; }
"&&"  	 { yylval.theop = ANDAND; add_text(); return ANDAND; }
"||"  	 { yylval.theop = OROR; add_text(); return OROR; }
"?"   	 { dprintf (1000, "QUEST"); add_text(); return yytext[0]; }
":"   	 { dprintf (1000, "COLON"); add_text(); return yytext[0]; }
","   	 { yylval.theop = COMMA; add_text(); return yytext[0]; }
";"   	 { dprintf (1000, "SEMI"); add_text(); return yytext[0]; }

{let}{alnum}*  {
  TraceRule("{let}{alnum}*");
  set_to_string (yylval.theconst, yytext);
  add_text(); 
  print_constant (1000, yylval.theconst); 
  return id_or_keyword_or_type (yytext);
}

{white}+ {
  TraceRule("{white}+");
  add_text();
}

^{white}+"#" {
  TraceRule("^{white}+\"#\"");
  add_text();
  printf ("pre command");
  return HASH;
}

{newline} {
  TraceRule("{newline}");
  add_text();
  LineNum++;
}

.     	       yyerror ("Illegal character <%s>\n", yytext);

%%

/********************************************************/

typedef struct {
   char *name;
   int val;
} KeyWord;

KeyWord KeyTab[] = {
   { "auto",   	  AUTO,       },
   { "break",  	  BREAK,      },
   { "case",   	  CASE,       },
   { "char",   	  CHAR,	      },
   { "continue",  CONTINUE,   },
   { "default",   DEFAULT,    },
   { "do",     	  DO,	      },
   { "double", 	  DOUBLE,     },
   { "else",   	  ELSE,	      },
   { "enum",   	  ENUM,	      },
   { "extern", 	  EXTERN,     },
   { "float",  	  FLOAT,      },
   { "for",    	  FOR,	      },
   { "goto",   	  GOTO,	      },
   { "if",     	  IF,	      },
   { "inline",    INLINE,     },
   { "int",    	  INT,	      },
   { "long",   	  LONG,	      },
   { "macro",  	  MACRO,      },
   { "pragma",    PRAGMA,     },
   { "register",  REGISTER,   },
   { "return", 	  RETURN,     },
   { "short",  	  SHORT,      },
   { "signed", 	  SIGNED,     },
   { "sizeof", 	  SIZEOF,     },
   { "static", 	  STATIC,     },
   { "struct", 	  STRUCT,     },
   { "switch", 	  SWITCH,     },
   { "typedef",	  TYPEDEF,    },
   { "union",  	  UNION,      },
   { "unsigned",  UNSIGNED,   },
   { "void",   	  VOID,	      },
   { "volatile",  VOLATILE,   },
   { "while",  	  WHILE,      }
};

int cmp (const void *va, const void *vb)
{
  KeyWord *a, *b;
  a = (KeyWord *)va;
  b = (KeyWord *)vb;
  return strcmp (a->name, b->name);
}

KeyWord* dumb_search( char* lex )
{
  int i, keytab_size;

  keytab_size = sizeof (KeyTab) / sizeof (KeyWord);

  for (i=0;i<keytab_size;++i)
    if ( strcmp(KeyTab[i].name,lex) >= 0 )
      break;
  if ( ( i == keytab_size )||( strcmp(KeyTab[i].name,lex) ) )
    return NULL;
  else
    return &(KeyTab[i]);
}

int id_or_keyword_or_type (char *lex)
{
  KeyWord *p, dummy;
  Type *t;
  
  dummy.name = lex;
  p = (KeyWord *)bsearch (&dummy, KeyTab, 
			  sizeof (KeyTab) / sizeof (KeyWord),
			  sizeof (KeyWord), 
			  cmp);
  if (p) {
    dprintf (1000, "KEYWORD: %s", p->name);
    if ( ( p->val == SIZEOF )||( p->val == RETURN ) )
      yylval.theop = p->val;
    return p->val;
  }
  else {
    if (t = find_type (lex)) {
      yylval.thetype = t;
      return TYPEDEFNAME;
    }
    else
      return IDENT;
    dprintf (1000, "IDENT: %s", lex);
  }
}

int yyerror(char *s)
{
  fprintf (stderr, "lex error in line: %d: %s: <%s>\n", LineNum, s, yytext);
}

void inityy(void)
{
# ifdef YYDEBUG
    yydebug = 1;
# endif
  SourceText = makebuf (128);
}

