/*
 * $RCSfile: usit_validate.c,v $
 *
 * Universal Stub Compiler
 *
 * Copyright (c) 1994    Arizona Board of Regents
 *
 * $Revision: 1.1 $
 * $Date: 1994/10/13 17:43:30 $
 */

#include "usc.h"
#include "usit.h"
#include "symbol.h"
#include "buffer.h"
#include "gen_utils.h"

static Buffer *ValidateDecls;
static Buffer *ValidateMain;

extern Buffer *FunctionsBuffer;

void val_gen_sym (char *prefix, Symbol *basesym, Symbol *sym);

void init_validate (void)
{
  ValidateDecls = makebuf (128);
  ValidateMain = makebuf (128);

  fprintf (UsitValidate, "#include <stdio.h>\n\n");
  writebuf (ValidateMain, "int main (void)\n{\n");
  writebuf (ValidateMain, "  unsigned long val_long, i, j;\n");
  writebuf (ValidateMain, "  int val_ret = 0;\n\n");
}

char *val_local_name (Symbol *sym)
{
  /*
  static char buf[1024];
  sprintf (buf, "val_%s", SYM_LOCAL_NAME (sym));

  return buf;
*/
  return SYM_LOCAL_NAME (sym);
}

void val_declare_sym (Symbol *sym)
{
  /* declare a symbol of this type in ValidateDecls */

  writebuf (ValidateDecls, "%s %s;\n", SYM_NAME (sym), val_local_name (sym));
}

int get_unique_label (void)
{
  static int label = 0;
  return label++;
}

void gen_val_code (char *name, Symbol *basesym, Type *the_type)
{
  int poke_label, peek_label;
  char *base_typename, *basename, *p, *q;
  char *param_name = NULL, *var_name = NULL, *sizeof_string;
  int param_len = 0, var_len = 0, sizeof_len;

  base_typename = SYM_NAME (basesym);
  basename = SYM_LOCAL_NAME (basesym);

  if (TYPE_IS_DECLARATOR (SYM_TYPE (basesym))) {
    if (SYM_IS_ARRAY (basesym)) {
      crash ("Talk to dorgival");
    }
    else {
      crash ("Some things just shouldn't be!\n");
    }
  }
  else {
    if (SYM_IS_STRUCT_OR_UNION (basesym)) {
      param_len = strlen (SYM_LOCAL_NAME (basesym)) + 1;
      param_name = (char *)Malloc ((param_len + 1) * sizeof (char));
      sprintf (param_name, "*%s", SYM_LOCAL_NAME (basesym));
      
      var_len = strlen (name) + 3;
      var_name = (char *)Malloc ((var_len + 1) * sizeof (char));

      q = var_name;
      p = name;
      *q++ = '(';
      *q++ = '*';
      while (*p && *p != '.')
	*q++ = *p++;
      *q++ = ')';
  
      while (*p)
	*q++ = *p++;
      *q = '\0';
    }
    else {
      param_len = strlen (SYM_LOCAL_NAME (basesym)) + 1;
      param_name = (char *)Malloc ((param_len + 1) * sizeof (char));
      sprintf (param_name, "*%s", SYM_LOCAL_NAME (basesym));
      
      var_len = strlen (SYM_LOCAL_NAME (basesym)) + 1;
      var_name = (char *)Malloc ((var_len + 1) * sizeof (char));
      sprintf (var_name, "*%s", SYM_LOCAL_NAME (basesym));
    }
  }

  if (TYPE_IS_BITFIELD (the_type)) {
    sizeof_len = 4;
    sizeof_string = (char *)Malloc ((sizeof_len + 1) * sizeof (char));
    sprintf (sizeof_string, "%d", TYPE_DECL_NUMEL (the_type));
  }
  else {
    /* base type */
    sizeof_len = sizeof ("8 * sizeof ()") + strlen (name);
    sizeof_string = (char *)Malloc ((sizeof_len + 1) * sizeof (char));
    sprintf (sizeof_string, "8 * sizeof (%s)", name);
  }
  
  poke_label = get_unique_label();
  writebuf (FunctionsBuffer, "void stub%d (%s %s, unsigned long *y)\n", 
	    poke_label, base_typename, param_name);
  writebuf (FunctionsBuffer, "{\n");
  writebuf (FunctionsBuffer, "  %s = *y;\n", var_name);
  writebuf (FunctionsBuffer, "}\n\n");
  
  peek_label = get_unique_label();
  writebuf (FunctionsBuffer, "void stub%d (unsigned long *y, %s %s)\n", 
	    peek_label, base_typename, param_name);
  writebuf (FunctionsBuffer, "{\n");
  writebuf (FunctionsBuffer, "  *y = %s;\n", var_name);
  writebuf (FunctionsBuffer, "}\n\n");
  
  /* now the test code */
  /* poke */
  
  writebuf (ValidateMain, "  for (i = 0; i < %s; i++) {\n",
	    sizeof_string, name);
  writebuf (ValidateMain, "    val_long =1L << i;\n");
  writebuf (ValidateMain, "    stub%d (&%s, &val_long);\n", 
	    poke_label, basename);
  writebuf (ValidateMain, "    for (j = 0; j < %s; j++) {\n", sizeof_string);
  writebuf (ValidateMain, "      if ((j == i && (%s & (1L << j)) != (1L << i))"
	    " || (j != i && (%s & (1L << j)) != 0)) {\n", 
	    name, name);
  writebuf (ValidateMain, "        fprintf (stderr, \"Poke failed: "
	    "%s %%x\\n\", 1L << i);\n", name);
  writebuf (ValidateMain, "        val_ret = 1;\n");
  writebuf (ValidateMain, "        break;\n");
  writebuf (ValidateMain, "      }\n    }\n  }\n");
  
  /* peek */
  
  writebuf (ValidateMain, "  for (i = 0; i < %s; i++) {\n",
	    sizeof_string, name);
  writebuf (ValidateMain, "    %s = 1L << i;\n", name);
  writebuf (ValidateMain, "    stub%d (&val_long, &%s);\n",
	    peek_label, basename);
  writebuf (ValidateMain, "    for (j = 0; j < %s; j++) {\n", sizeof_string);
  writebuf (ValidateMain, "      if ((j == i && (val_long & (1L << j)) != "
	    "(1L << i)) || (j != i && (val_long & (1L << j)) != 0)) {\n");
  writebuf (ValidateMain, "        fprintf (stderr, \"Peek failed: "
	    "%s %%x\\n\", 1L << i);\n", name);
  writebuf (ValidateMain, "        val_ret = 1;\n");
  writebuf (ValidateMain, "        break;\n");
  writebuf (ValidateMain, "      }\n    }\n  }\n");
  

  dprintf (0, "Validating: %s peek: stub%d poke: stub%d", name, peek_label,
	   poke_label);

  Free (param_name, param_len);
  Free (var_name, var_len);
  Free (sizeof_string, sizeof_len);
}
  

void val_gen_spec (char *prefix, Symbol *basesym, Type *spec)
{
  char *string;
  int string_len; 
  Symbol *field;

  if (TYPE_IS_STRUCT_OR_UNION (spec)) {
    string_len = strlen (prefix) + 1;
    string = (char *)Malloc ((string_len + 1) * sizeof (char));
    sprintf (string, "%s.", prefix);
    
    for (field = TYPE_STRUCTDEF_FIELDS (spec); field;
	 field = SYM_NEXT (field)) {
      val_gen_sym (string, basesym, field);
    }
    
    Free (string, string_len + 1);
  }
  else {
    gen_val_code (prefix, basesym, spec);
  }
}

void val_gen_array (char *prefix, Symbol *basesym, Type *the_type)
{
  char *string, buf[256];
  int string_len;
  int i, num_el;

  if (TYPE_IS_DECLARATOR (the_type)) {
    num_el = TYPE_DECL_NUMEL (the_type);
    sprintf (buf, "%d", num_el);

    string_len = 2 + strlen (buf) + strlen (prefix);
    string = (char *)Malloc ((string_len + 1) * sizeof (char));

    for (i = 0; i < num_el; i++) {
      sprintf (string, "%s[%d]", prefix, i);
      val_gen_array (string, basesym, TYPE_NEXT (the_type));
    }

    Free (string, string_len + 1);
  }
  else {
    val_gen_spec (prefix, basesym, the_type);
  }
}

void val_gen_sym (char *prefix, Symbol *basesym, Symbol *sym)
{
  /* walk the symbol tree, making lots of stubs and  test code in parallel */

  char *name, *temp;
  int name_len;

  if (*prefix)
    temp = SYM_NAME (sym);
  else
    temp = SYM_LOCAL_NAME (sym);

  name_len = strlen (prefix) + strlen (temp);
  name = (char *)Malloc ((name_len + 1) * sizeof (char));
  sprintf (name, "%s%s", prefix, temp);

  if (TYPE_IS_SPECIFIER (SYM_TYPE (sym))) {
    val_gen_spec (name, basesym, SYM_TYPE (sym));
  }
  else if (SYM_IS_ARRAY (sym)) {
    val_gen_array (name, basesym, SYM_TYPE (sym));
  }
  else if (SYM_IS_BITFIELD (sym)) {
    gen_val_code (name, basesym, SYM_TYPE (sym));
  }
  
  Free (name, name_len + 1);
}

void usit_validate_sym (Symbol *sym)
{
  val_declare_sym (sym);
  val_gen_sym ("", sym, sym);
}

void gen_usit_validate (UsitDecList *the_list)
{
  UsitDecl *p;

  for (p = the_list->head; p; p = p->next) {
    usit_validate_sym (p->the_sym);
  }

  writebuf (ValidateMain, "  return val_ret;\n}\n");

  if (BUFFER_SIZE (ValidateDecls) > 0)
    fprintf (UsitValidate, "%s\n", BUFFER_STRING (ValidateDecls));
  fprintf (UsitValidate, "%s\n", BUFFER_STRING (ValidateMain));
}
